	list      p=16c774            ; list directive to define processor
	#include <p16c774.inc>
;	Definitions for PIC16F877 
; processor specific variable definitions
	#include <xbanks.inc>
; macros for automatic bank handling
 
	__CONFIG   _CP_OFF & _WDT_OFF & _BODEN_ON & _PWRTE_ON & _HS_OSC 
	NOEXPAND

;***** VARIABLE DEFINITIONS
tx_buf_in_ptr equ 0x20		; tail of transmitter buffer
tx_buf_out_ptr equ 0x21		; head of transmitter buffer
rs_chksum     equ 0x22		; ???
chan_ptr equ 0x23		; Pointer on current channel
chan_num_act equ 0x24		; Number of actually transmitted channels
chan_num equ 0x25		; Number of sampled channels
main_tmp equ 0x26		; temporary variable used in main thread
int_tmp equ 0x27		; temporary variable used in interrupt context
rx_buf_in_ptr equ 0x28		; tail of receiver buffer
rx_buf_out_ptr equ 0x29		; head of receiver buffer
UserFlags1 equ 0x2a		; User bit flags
; User bit flags definitions
UF1_rx_overflow equ 0		; receiver buffer overflow
UF1_STOP equ 1			; stop sampling after the record is completed
UF1_SAMPLE equ 2		; sampling is on
UF1_WAS_DIN equ 3		; digital inputs have been read for this channel
	
msg_ptr	equ 0x2b		; variable used for sending of text messages
frq_r1 equ 0x2c			; variables used for setting of sampling frequency
frq_r2 equ 0x2d
frq_r3 equ 0x2e
record_number equ 0x2f		; number of the record
prev_chan equ 0x30		; number of previous channel
din_pb equ 0x31			; values read from port B
din_pd equ 0x32			; values read from port D
tmp_rcb equ 0x33		; temporary variable used in receive interrupt
w_temp        EQU     0x70       ; variable used for context saving 
status_temp   EQU     0x71       ; variable used for context saving
fsr_temp      EQU     0x72       ; variable used for context saving
adc_tmp1      equ     0xa0	; temporary variable used in analog input interrupt
tx_buf_start equ 0x120		; first byte of transmitter buffer
rx_buf_start equ 0x140		; first byte of receiver buffer
chan_buf equ 0x160		; first byte of channels buffer

;***** CONSTANTS DEFINITIONS
tx_buf_mask equ 0x1f		; mask for calculation of position in transmitter buffer
rx_buf_mask equ 0x1f		; mask for calculation of position in receiver buffer
CHAN_NUM_MAX equ 0x10		; maximum number of channels

;***** SPECIAL COMMENDS RECEIVED VIA RS232 INTERFACE
RSC_RESET equ 0xff		; reset the converter

;***** RESPONSES SENT VIA RS232 TO PC
RSR_TOO_FAST equ 0xfe		; sampling was too fast
RSR_UNK_CMD equ 0xfd		; unknown command
RSR_TOO_MANY_CHAN equ 0xfc	; too many channels requested
RSR_RXBUF_OVF equ 0xfb		; receiver buffer overflow

RSR_OK equ 0xc0			; positive acknowledge
RSR_INIT equ 0xc1		; converter initialized

;***** HELPER MACROS
;Update of checksum (W contains the transferred byte)
upd_chksum macro
		xaddwf	rs_chksum,f
		btfsc	rs_chksum,7
		incf	rs_chksum,f
		bcf	rs_chksum,7
		endm		

; Macros used for commands decoding
; PIC does not offer the COMPARE instruction, therefore the XOR
; instruction is used to decode the commands.
; However this instruction modifies the contents of W.
; The previous modifications are considered when checking each code
wzcmp_init macro
	variable wzcmp_prev_val=0 ; this variable stores modifications - initially 0
	endm

wzcmp macro test_val
 	xorlw wzcmp_prev_val ^ test_val ; XOR with checked code and previous modification
	variable wzcmp_prev_val = test_val ; store the current modification
	endm

;**********************************************************************
		ORG     0x000           ; processor reset vector
		clrf    PCLATH          ; ensure page bits are cleared
  		goto    main            ; go to beginning of program


		ORG     0x004           ; interrupt vector location
		goto	irq_svc		; jump to the interrupt service routine

; The procedure below displays the initial message
welc_msg
	addwf PCL,f
	dt RSR_INIT
	dt "PICADC 3.0.8",0xd
	dt "4608000",0xd ; fxtal/4=18432000/4 - base frequency when using the PIC oscillator
	dt "16000000",0xd ; Base frequency when using the additional oscillator
	dt "9",0xd ; Number of available channels
	dt "16",0xd,0x0 ; Maximal length of channels' queue

; The macro translating the number of analog input channel into the value
; to be written into ADCON0
adc_num_gen macro kan_num
	variable adc_val = B'00000001'
	if kan_num & 1
	  variable adc_val = adc_val | 0x08
	endif
	if kan_num & 2
	  variable adc_val = adc_val | 0x10
	endif
	if kan_num & 4
	  variable adc_val = adc_val | 0x20
	endif
	if kan_num & 8
	  variable adc_val = adc_val | 0x02
	endif
	; The bit 0x80 has a special meaning - it describes that the digital
	; inputs are to be read together with that channel
	if kan_num & 0x80
	  variable adc_val = adc_val | 0x80
	endif
	dt	adc_val
	endm

; The procedura below translates the number of channels into the values written into
; ADCON0 to trigger sampling in the particular channel
; Uses the previously defined macro
adc_chan_translate
	andlw 0x1f  ; Channel number must not be higher than 0x1f
	addwf PCL,f
	adc_num_gen	0x0	; 0
	adc_num_gen	0x1     ; 1
	;adc_num_gen	0x2 The channel 2 is not used as input. This pin provides 2.048V reference
	adc_num_gen	0x3     ; 2 
	adc_num_gen	0x4     ; 3
	adc_num_gen	0x5     ; 4
	adc_num_gen	0x6     ; 5
	adc_num_gen	0x7     ; 6
	adc_num_gen	0x8     ; 7
	adc_num_gen	0x9     ; 8
	adc_num_gen	0x0	; 9 Starting from this line we repeat first channels
	adc_num_gen	0x1	; a so that we have 16 channels and there are no 
	adc_num_gen	0x3	; b invalid channel numbers
	adc_num_gen	0x4     ; c
	adc_num_gen	0x5     ; d
	adc_num_gen	0x6     ; e
	adc_num_gen	0x7     ; f
	; Now the channels with digital input
	adc_num_gen	0x80	; 10
	adc_num_gen	0x81    ; 11
	;adc_num_gen	0x2 The channel 2 is not used as input. This pin provides 2.048V reference
	adc_num_gen	0x83    ; 12 
	adc_num_gen	0x84    ; 13
	adc_num_gen	0x85    ; 14
	adc_num_gen	0x86    ; 15
	adc_num_gen	0x87    ; 16
	adc_num_gen	0x88    ; 17
	adc_num_gen	0x89    ; 18
	adc_num_gen	0x80	; 9 Starting from this line we repeat first channels
	adc_num_gen	0x81	; a so that we have 16 channels and there are no 
	adc_num_gen	0x83	; b invalid channel numbers
	adc_num_gen	0x84    ; 1c
	adc_num_gen	0x85    ; 1d
	adc_num_gen	0x86    ; 1e
	adc_num_gen	0x87    ; 1f


main
	xbank_inv
	; We switch the interrupts off 
	xclrf PIE1
	xclrf PIE2
	; We program the pin data directions
	; Analog inputs
	; RA0-RA3 = AN0-AN3 
	; RA5 = AN4
	; RE0-RE2 = AN5-AN7
	; RB2-RB3 = AN8-AN9
	movlw	B'00101111'
	xmovwf	TRISA
	movlw	B'10001100' ; Pin B7 - reset, pins 1 and 4-6 - digital I/O
	xmovwf	TRISB
	; Port C will be set, when programming the serial port
	; Port E
	movlw	B'00000111'
	xmovwf	TRISE
	; Port D will be used as the digital inputs
	movlw	0xff
	xmovwf	TRISD
	; We set up the voltage reference
	movlw	(1<<VRLEN) | (1<<VRLOEN) | (1<<VRHEN)
	xmovwf	REFCON
	; We set up the A/D converter
	; Left adjusted result, reference 4.096-GND, all analog pins
	movlw	(0<<ADFM)|(B'100'<<VCFG0)|(B'0000'<<PCFG0)
	xmovwf	ADCON1
	; The initial contents of channels table (mainly for testing with GPSIM,
	; or for the situation when someone starts the sampling without
	; setting the channels first 
	movlw   0x01
	xmovwf	chan_buf
	xmovwf	(chan_buf+1)
	xmovwf	(chan_buf+2)
	xmovwf	(chan_buf+3)
	xmovwf	(chan_buf+4)
	xmovwf	(chan_buf+5)
	movlw   0x05
	xmovwf  chan_num
	movlw	0x03
	xmovwf	chan_num_act
	; We initialize the variables used for cyclic buffers
	; the serial port buffers
	xclrf   tx_buf_in_ptr
	xclrf   tx_buf_out_ptr
	xclrf	rx_buf_in_ptr
	xclrf   rx_buf_out_ptr
	; the channels' buffer
	xclrf	chan_ptr
	; clear the user flags
	xclrf	UserFlags1
	; We initialize the hardware necessary to communicate with PC
	; Set up the serial port and program the transmission rate
	movlw   B'11000001' ; C7,C6 - set to inputs to enable serial port 
			    ; C0 - additional clock input
	xmovwf	TRISC
	movlw   0x09 ; Baud rate 115200, xtal 18432000
;	movlw   0x04 ; Baud rate 230400, xtal 18432000
	xmovwf  SPBRG
	movlw	(1<<BRGH)
	xmovwf	TXSTA
	movlw   (1<<SPEN)|(1<<CREN)
	xmovwf  RCSTA
	xbsf	PIE1,TXIE
	xbsf	PIE1,RCIE	
	movlw	(1<<TXEN)|(1<<BRGH)
	xmovwf	TXSTA
	; Now the serial port is ready
	; Switch on the serial port interrupts
	xbsf	INTCON,GIE
	xbsf	INTCON,PEIE
	; Now we can send the initial message
cmd_inicjalizacja xbank_inv
wlc_msg		xbank_inv
	xclrf	msg_ptr	
wlc_msg1	xbank_inv
	xmovfw  msg_ptr
	call	welc_msg
	iorlw	0
	btfsc   STATUS,Z
	goto	wlc_end
	call	tx_send
	xbank_inv
	btfsc   STATUS,Z
        goto    wlc_msg1
	xincf	msg_ptr,f
	goto	wlc_msg1
wlc_end		xbank_inv
; After the initial message is sent, we wait until PC sends a command
main_loop1	xbank_inv
	call	rx_read_wait
	; We received a character
main_rcvd	wzcmp_init
	wzcmp	"R" ; Restart (redundant command, 0xff does the same more reliably)
	btfsc	STATUS,Z
	goto	0
	wzcmp	"I" ; Initialization of the hardware
	btfsc	STATUS,Z
	goto	cmd_inicjalizacja
	wzcmp	"C" ; Initialization of the channels' table
	btfsc	STATUS,Z
	goto	cmd_kanaly
	wzcmp	"F" ; Initialization of the frequency divider
	btfsc	STATUS,Z
	goto	cmd_czestotliwosc
	wzcmp	"P" ; Start sampling
	btfsc	STATUS,Z
	goto	cmd_start_probk
	; If we get here - the unknown command has been received
	; We send the negative acknowledge and continue
        wzcmp   0 ; Restore the real code of unknown command and send it to PC
	call	tx_send_wait
	movlw	RSR_UNK_CMD ; Unknown command notification
	call	tx_send_wait
	goto	main_loop1

cmd_czestotliwosc xbank_inv
	movlw	"F"
	call	tx_send_wait
	xbank_inv
	xclrf	rs_chksum
	; We read three 7-bit words
	call	rx_read_wait
	call	tx_send_wait
	xbank_inv
	upd_chksum
	xmovwf	frq_r1
	call	rx_read_wait
	call	tx_send_wait
	xbank_inv
	upd_chksum
	xmovwf	frq_r2
	call	rx_read_wait
	call	tx_send_wait
	xbank_inv
	upd_chksum
	xmovwf	frq_r3
	; We send the checksum
	xmovfw  rs_chksum
	call	tx_send_wait		
	; We build the divider value
	xbank_inv
	; Shift the LSB of frq_r2 into MSB of frq_r1
	xrrf	frq_r2,w
	btfsc   STATUS,C
	xbsf	frq_r1,7
	andlw	0x3f
	movwf	frq_r2
	; Shift bits 0&1 of frq_r3 into bits 6&7 of frq_r2
	xrrf	frq_r3,f
	btfsc	STATUS,C
	xbsf	frq_r2,6
	xrrf	frq_r3,w
	btfsc	STATUS,C
	xbsf	frq_r2,7
	andlw	0x1f
	movwf	frq_r3
	goto	main_loop1
;
; Read the input channels' list
cmd_kanaly	xbank_inv
	; We send the positive acknowledge
	movlw	"C"
	call	tx_send_wait
	; First we fill the whole array with the code of channel 0
	xbank_inv
	movlw	chan_buf & 0xff
	xmovwf	FSR
	movlw	CHAN_NUM_MAX
	xmovwf	chan_num
	movlw	0
	call	adc_chan_translate ; W register holds the channel 0 code
	xbank_inv
	bsf	STATUS,IRP
	xbank_vset chan_num
	; IMPORTANT!!! the section below (until "goto") must use the same bank!!!
cmd_kan_fill1   movwf	INDF
	incf	FSR,f
	decfsz	chan_num,f
	goto	cmd_kan_fill1
	; end of critical section
	;
	; We read the input channels
	; We clear the input channel pointer
	xclrf	chan_num
	xclrf	chan_ptr
	; We clear the checksum
	xclrf	rs_chksum		
	; First we receive the amount of really sampled channels
cmd_kan_loop0   xbank_inv
	call	rx_read_wait
	xbank_inv
	xmovwf	chan_num_act
	upd_chksum
	call	tx_send_wait
cmd_kan_loop1	xbank_inv
	call	rx_read_wait
	xbank_inv
	xmovwf	main_tmp
	xorlw	0x7f  ; Marker of the end of channels' table
	btfsc	STATUS,Z
	goto	cmd_kan_end
	; If it was not the end marker, we check if the amount of input channels
	; has not been exceeded
	xmovfw	chan_num
	xorlw	CHAN_NUM_MAX
	btfsc	STATUS,Z
	goto	cmd_kan_too_many
	; So we must have a number of the next channel
        xmovfw	main_tmp		
	upd_chksum
	call	tx_send_wait
	; We translate the number of the channel into the value to be written
	; into the ADCON0
	call	adc_chan_translate
	xbank_inv				
	; We calculate the address where the new channel is to be stored
	xmovwf	main_tmp
	bsf	STATUS,IRP
	movlw	(chan_buf & 0xff)
	xaddwf	chan_num,w
	movwf	FSR
	xmovfw	main_tmp
	movwf	INDF
	; New channel stored
	; Increase the number of the channel
cmd_brk	xincf	chan_num,f
	; Return to the begining of the routine
	goto	cmd_kan_loop1
cmd_kan_too_many
	xbank_inv			
	; Inform PC, that the amount of channels has been exceeded	
	movlw	RSR_TOO_MANY_CHAN
	call	tx_send_wait
	goto	main_loop1
cmd_kan_end
	xbank_inv
	; We echo the marker of table's end
	movlw	0x7f
	upd_chksum
	call	tx_send_wait
	; We send the checksum as a confirmation
	xbank_inv
	xmovfw	rs_chksum
	call	tx_send_wait
	goto	main_loop1

cmd_start_probk xbank_inv
	; Send the confirmation of the command 
	movlw	"P"
	call	tx_send_wait
	xbank_inv
	; We set the CCP frequency
	xmovfw	frq_r2
	xmovwf	CCPR2H
	xmovfw	frq_r1
	xmovwf	CCPR2L
	; We set the "trigger special event" mode in CCP
	movlw 	0x0b
	xmovwf	CCP2CON
	xclrf	TMR1L
	xclrf	TMR1H
        ; In main_tmp we create the timer1 control word
	xclrf	main_tmp
	xbtfsc	frq_r3,0
	xbsf	main_tmp,T1CKPS0
	xbtfsc	frq_r3,1
	xbsf	main_tmp,T1CKPS1
	xbtfsc	frq_r3,2
	xbsf	main_tmp,TMR1CS
	xbtfsc	frq_r3,3
	xbsf	main_tmp,T1SYNC
	xbsf	main_tmp,TMR1ON
	xmovfw	main_tmp
	xmovwf	T1CON
	; Send the header of the first record
	xclrf	record_number
	movlw	0x80
	call	tx_send
	xbank_inv
	xclrf	rs_chksum
	; We set the parameters so that the first sampling will be done in first channel
	xmovfw	chan_buf
	xclrf	chan_ptr
	xmovwf  prev_chan       ; store the channel number to check if we should read the
				; digital input 
	iorlw   0x80		; the MSB must be set to select 1/32 fx conversion clock
	xmovwf	ADCON0
	; Now we unlock the A/D converter
	xbcf	PIR1,ADIF
	xbsf	PIE1,ADIE
	xbsf	UserFlags1,UF1_SAMPLE
; Below is the sampling loop. All commands except "S" (end of sampling)
; and 0xff (reset) are ignored
probk_loop1	xbank_inv
	call	rx_read_wait
	xbank_inv
	xorlw	"S"
	btfsc	STATUS,Z
	goto	probk_loop1
	xbsf	UserFlags1,UF1_STOP ; We stop sampling after the last record is finished
	; Wait until the end of last record is confirmed
probk_loop2	xbank_inv
	xbtfsc  UserFlags1,UF1_STOP
	goto	probk_loop2
	; We send the positive acknowledgment
	movlw	RSR_OK 
	call	tx_send_wait
	goto	main_loop1	; Return to the main loop

; INTERRUPT ROUTINES
; Serial port receiver interrupt
rx_int_handler
	xbank_inv
	; First we check if it is a special code (reset or digital output setting)
	xmovfw RCREG
	xmovwf tmp_rcb	; we store the received character 
	xbtfss tmp_rcb,7 ; if the MSB is not set, this is not a special command  
	goto  rx_int_normal
	; special code received 
	; register W still holds the received character
	xorlw	RSC_RESET
	btfsc	STATUS,Z
	goto	soft_reset	; RESET command - go to reset
	; otherwise it is the value to be send to digital outputs
	; write it directly to the PORTB (keep in mind that all bits are
	; inverted after comparison to RSC_RESET !!!)
	xmovwf  PORTB
	return
rx_int_normal: xbank_inv
	xincf rx_buf_in_ptr,w	; First check if there is free space in receive buffer
	andlw rx_buf_mask
	movwf FSR
	xxorwf rx_buf_out_ptr,w
	btfsc STATUS,Z
	goto rx_buf_overflow	; The receiver buffer is overflown 
	movfw FSR
	xmovwf rx_buf_in_ptr	; update buffer pointer
	addlw rx_buf_start & 0xff ; We calculate the pointer to the free place in buffer
	bsf	STATUS,IRP
	movwf FSR
	xmovfw tmp_rcb		; get the received character from the temporary variable
	xmovwf INDF		; store it to the receive buffer
	return
rx_buf_overflow
	xbank_inv
	movlw	RSR_RXBUF_OVF
	goto	send_alarm_wait_reset

; The routine below receives the character from the serial port
; The procedure does not wait
; Z flag set - no byte was available, Z flag cleared - the byte has been received
rx_read		xbank_inv
	; Check if the buffer is empty
	xmovfw	rx_buf_in_ptr
	xxorwf  rx_buf_out_ptr,w
	btfsc   STATUS,Z
	goto    rx_read_return ; buffer was empty
	xincf   rx_buf_out_ptr,w
	andlw   rx_buf_mask
	xmovwf  rx_buf_out_ptr	; update the buffer pointer
	addlw   rx_buf_start & 0xff
	bsf	STATUS,IRP
	movwf   FSR
	xmovfw  INDF
	bcf	STATUS,Z
rx_read_return  return  						

; The waiting read character routine
rx_read_wait	xbank_inv
	call	rx_read
	btfsc	STATUS,Z
	goto	rx_read_wait
	return

; The macro below puts the byte into the transmit buffer
rs_buf_write macro byte_to_send
	xbank_inv
	; first check if there is a free space in the buffer
	xincf tx_buf_in_ptr,w
	andlw tx_buf_mask ; mask the higher bits 
	xmovwf tx_buf_in_ptr
	xmovwf FSR
	xxorwf tx_buf_out_ptr,w
	btfsc STATUS,Z
	; The transmit buffer has been overflown ;
	; All we can do is to send the warning and stop sampling
	goto tx_buf_overflow
	bsf STATUS,IRP
	movlw tx_buf_start & 0xff
	addwf FSR,f
        xmovfw byte_to_send
	movwf INDF
	upd_chksum		; update the checksum
	xbsf  PIE1,TXIE		; turn on the transmitter interrupt
	xbank_inv
	endm
				
; End of conversion interrupt
adc_int_handler
	xbank_inv
	xbtfss	UserFlags1,UF1_SAMPLE ; If UF1_SAMPLE flag is cleared, ignore the interrupt
	return
	; First we read the digital input if it is required
	xbcf    UserFlags1,UF1_WAS_DIN
	xbtfss  prev_chan,7
	goto    adc_int_no_digital
	xmovfw  PORTB
	xmovwf  din_pb
	xmovfw  PORTD
	xmovwf  din_pd
	xbsf    UserFlags1,UF1_WAS_DIN
adc_int_no_digital xbank_inv
	; now we set the analog mux to the next channel
	xmovfw	chan_ptr
	xmovwf	int_tmp
	xincf  chan_ptr,f
	xmovfw	chan_num
	xsubwf	chan_ptr,w	; This instruction sets the bank for chan_ptr
	btfsc	STATUS,C
	clrf	chan_ptr ; The previous instruction didn't modify the bank setting
	xmovfw	chan_ptr
	addlw chan_buf & 0xff
	bsf STATUS,IRP
	movwf  FSR
	movfw  INDF
	xmovwf  prev_chan
	iorlw   0x80		; the MSB must be set to select 1/32 fx conversion clock
	xmovwf  ADCON0
	; Receive the data from the converter
	bcf STATUS,C 
	xrrf ADRESH,w
	xmovwf adc_tmp1
	xrrf ADRESL,f
	bcf STATUS,C
	xrrf adc_tmp1,f
	xrrf ADRESL,f
	bcf STATUS,C
	xrrf ADRESL,f
	bcf STATUS,C
	xrrf ADRESL,f
	; The conversion result has been packed so that only 6 bits are used
	; in each byte
	; Now we check if it is end of the record
adc_brk2 xmovfw chan_num_act
	xsubwf int_tmp,w
	btfsc STATUS,C
	goto adc_int_dont_transm ; Check if the data should be transmitted
	rs_buf_write ADRESL
	rs_buf_write adc_tmp1
	; Now we check if we need to write the digital input
	xbtfss UserFlags1,UF1_WAS_DIN
	goto adc_int_dont_transm
	; We should process and transfer the data from digital input
	; The data are stored as follows:
	; D4-D11 in din_pd.0-7, D0-D2 in din_pb.4-6 and D3 in din_pb.1
	; First move D3 to din_pb.7
	xbcf din_pb,7
	btfsc din_pb,1
	bsf din_pb,7	; It is essential, that previous instruction set the bank to din_pb!
	; Now we move D4 and D5 to din_pb, shifting D0-D3 down
	; Simultaneously we shift "01" to the din_pd.7-6
	bsf STATUS,C
	xrrf din_pd,f
	xrrf din_pb,f
	bcf STATUS,C
	xrrf din_pd,f
	xrrf din_pb,f
	; Now we need to shift "01" to the din_pb.7-6
	bsf STATUS,C
	xrrf din_pb,f
	bcf STATUS,C
	xrrf din_pb,f
	; Finally send the processed data 
	rs_buf_write din_pb
	rs_buf_write din_pd
adc_int_dont_transm xbank_inv
	xmovfw chan_ptr
	btfss STATUS,Z
	goto adc_int_return
adc_int_chksum xbank_inv
	rs_buf_write rs_chksum
	; Now we should check if there is "stop sampling" request
	xbtfsc UserFlags1,UF1_STOP
	goto adc_int_stop
	; If there is no stop request - send header of the next record
	xincf record_number,f
	xbsf  record_number,7
	xbcf  record_number,6
	rs_buf_write record_number
	xclrf rs_chksum
adc_int_return	
	xbank_inv	
	return

adc_int_stop ; Stop sampling request routine
	xbank_inv
	xbcf	UserFlags1,UF1_STOP
	xbcf	UserFlags1,UF1_SAMPLE
	xclrf	T1CON ; stop TIMER1
	return

tx_buf_overflow 
	xbank_inv
	; This routine is called from the interrupt context 
	; therefore we can poll TXIF and RCIF, ignoring
	; the buffers
	; When the transmitter is free, we send the error message
	; and wait for RESET request
	movlw	RSR_TOO_FAST
send_alarm_wait_reset
	xbank_inv
	xbtfss	PIR1,TXIF
	goto	send_alarm_wait_reset
	xmovwf	TXREG
	; Error notification sent, wait for RESET
	; Normally the RESET request is serviced in the interrupt routine
	; but now we are already in the interrupt context
sawr_loop1
	xbank_inv
	xbtfss	PIR1,RCIF
	goto	sawr_loop1
	xmovfw	RCREG
	xorlw	RSC_RESET ; Reset request
	btfss	STATUS,Z
	goto	sawr_loop1
	; We continue with the reset procedure
; Procedure resetting the microcontroller
soft_reset
	xbank_inv
	bcf	INTCON,GIE ; Switch off the interrupts so nothing could 
	movlw	B'11111111' ; Set all port B pins to inputs
	xmovwf  TRISB
	xclrf	PORTB	    ; Set port B latch to 0
	movlw	B'01111111' ; Set B7 bit to output - B7 is connected to the MCLR pin
		 	    ; so the ucontroller will get reinitialized
	xmovwf	TRISB
	xclrf	PORTB	    ; Set port B latch to 0
	goto $ 	; The last two instructions probably will not execute, but anyway 
		; we protect our code with the endless loop

; The transmitter interrupt routine
tx_int_handler
	xbank_inv
	xmovfw  tx_buf_out_ptr
	xxorwf  tx_buf_in_ptr,w
	btfsc   STATUS,Z	; check if there are any data to transmit
	goto    tx_int_nodata
	; There are data to be transmited
        xincf  tx_buf_out_ptr,w
	andlw  tx_buf_mask
	xmovwf tx_buf_out_ptr
	addlw  tx_buf_start & 0xff
	bsf    STATUS,IRP
	movwf  FSR
	movfw  INDF
	xmovwf TXREG
tx_brk	goto   tx_int_return
tx_int_nodata
        xbank_inv
	; because there are no data, we block transmitter interrupts
	; (they will be unlocked by the transmit routine)
        xbcf  PIE1,TXIE	
tx_int_return	
        xbank_inv
	return

; The procedure for sending the data via the serial port
; Z flag set - data was not send because the buffer was full
; Z flag cleared - data has been written into transmit buffer
; IMPORTANT!!! Does not modify the W register
tx_send xbank_inv
	bcf	INTCON,GIE ; Block the interrupts to perform pointer operations
	xmovwf	main_tmp
	; first check if there is free space in the buffer
	xincf	tx_buf_in_ptr,w
	andlw	tx_buf_mask ; wrap the pointer
	xxorwf	tx_buf_out_ptr,w
	btfsc	STATUS,Z
	goto	tx_send_return	; the transmit buffer overflown - ignore the data
	;
	xincf 	tx_buf_in_ptr,w
	andlw	tx_buf_mask ; wrap the pointer
	movwf	tx_buf_in_ptr
	bsf	STATUS,IRP
	addlw	tx_buf_start & 0xff
	movwf	FSR
	xmovfw	main_tmp
	xmovwf	INDF
	bcf	STATUS,Z
tx_send_return
	xbank_inv
	bsf	INTCON,GIE ; unlock interrupts
	xbsf	PIE1,TXIE
	xbank_inv
	return

; The procedure below sends the charracter via the serial port,
; waiting until it is successfully written into the transmit buffer
; (does not wait until it is really transmitted)
tx_send_wait	xbank_inv
	call	tx_send
	btfsc	STATUS,Z
	goto	tx_send_wait
	return

; Entry point for different interrupt routines
irq_svc	xbank_inv
	movwf   w_temp            ; save the current W register contents
	movf	STATUS,w          ; move status register into W register
	movwf	status_temp       ; save the contents of STATUS register
	movfw	FSR
	movwf	fsr_temp

; first we check if it is serial port rx interrupt (most critical)
int_test_rx_irq
	xbank_inv
	xbtfss	PIR1,RCIF	; is it rx interrupt?
	goto	int_test_tx_irq ; nope check next
	call	rx_int_handler	; goto rx handler
        xbank_inv
int_test_tx_irq
	xbank_inv
	xbtfss	PIR1,TXIF	; is it tx interrupt?
	goto	int_test_adc_irq
	call	tx_int_handler
	xbank_inv
int_test_adc_irq
	xbank_inv
	xbtfss	PIR1,ADIF	; is it ADC interrupt?
	goto	INT_TEST_NXT
	call	adc_int_handler
	xbank_inv
	xbcf	PIR1,ADIF       ; clear the ADC interrupt flag
INT_TEST_NXT
	xbank_inv
	movfw	fsr_temp
	movwf	FSR
	movf    status_temp,w     ; retrieve copy of STATUS register
	movwf	STATUS            ; restore pre-isr STATUS register contents
	swapf   w_temp,f
	swapf   w_temp,w          ; restore pre-isr W register contents
	retfie                    ; return from interrupt

	END                       ; directive 'end of program'



